/*
 * drive.c - drive WS2812B
 *   by uratan! 2014.10.18
 */
#include "common.h"
#include <chip.h>

/*
 * I will drive PIO0_3
 */
#define PIO0_n_DIN       3	/* use PIO0_3 */
#define BITMAP_DIN   (1<<PIO0_n_DIN)

/*
 * functions
 */
void drive_c_init(u_long sysCLK);
void drv_select_method_GPIO(void);
void drv_select_method_SPI(void);
void drv_send_code(u_long *m, int N_leds);
void drv_GPIO_send_code(u_long *m, int N_leds);
void drv_GPIO_make_a_pulse(int tt1, int tt2, int bitmap, VOLATILE u_long *reg);
void drv_SPI_send_code(u_long *m, int N_leds);

/*
 * variables
 */
int drv_use_method_SPI;
int drv_GPIO_pulse_wide;
int drv_GPIO_pulse_narrow;


/* =============================================================== */

/*
 * initial misc for both SPI and GPIO
 */
void drive_c_init(u_long sysCLK)
{
	u_long n, tmp;

	/*
	 * initial mine
	 */
	drv_GPIO_pulse_wide   = 6;
	drv_GPIO_pulse_narrow = 3;
	drv_use_method_SPI = TRUE;

	/*
	 * initial as GPIO
	 *   set default output level at first
	 */
	LPC_GPIO_PORT->CLR[0] = BITMAP_DIN;		/* set Low */
	LPC_GPIO_PORT->DIR[0] |= BITMAP_DIN;		/* start to drive */

	/*
	 * initial as SPI
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_SPI0);

	LPC_SPI0->INTENCLR = 0x3f;		/* do not use interrupts */
	LPC_SPI0->DIV = sysCLK / (3*1000*1000) - 1;	/* SCK: 3MHz */
	LPC_SPI0->DLY = 0;				/* no delays are needed */
	LPC_SPI0->TXCTRL = (1<<22)			/* ignore receive */
					 | ((15-1) <<24)	/* 15bit/frame */
					 | 0;
	LPC_SPI0->CFG = (1<<0)		/* ENABLE */
				  | (1<<2)		/* MASTER mode */
				  | (0<<3)		/* MSB first */
				  | (1<<4)		/* CPHA */
				  | (1<<5)		/* CPOL */
				  | 0;

	/*
	 * different from SSP peripheral of LPC1830,
	 * first bit of 'the transfer' leaks before SSEL goes low.
	 *   (at LPC1830, the chagne of MOSI is completely sync with SCK)
	 * so I can't control the width of first bit.
	 *
	 * but in the middle of 'the frame', MOSI changes completely sync
	 * with SCK, so I will use never-ending-frame to drive.
	 *   (never use EOT/EOF conditions, keep 'in-frame' state)
	 */
	while(! (LPC_SPI0->STAT & SPI_STAT_TXRDY) ) {
		;
	}
	LPC_SPI0->TXDAT = 0x0000;
		/*
		 * by this initial transfer, SSEL goes low and 0x0000 is trasfered.
		 * the SPI peripheral is in transfering-frame state, keep this.
		 *   (SSEL will never be negated)
		 */

	/*
	 * set default method as thru_SPI
	 */
	if(drv_use_method_SPI) {
		drv_select_method_SPI();
	} else {
		drv_select_method_GPIO();
	}

	/*
	 * finally, disable pull-up-R on output port (PIO0_3)
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

	n = IOCON_PIO3;
	tmp = LPC_IOCON->PIO0[n];
	tmp &= ~PIN_MODE_MASK;
	tmp |= (PIN_MODE_INACTIVE <<PIN_MODE_BITNUM);
	LPC_IOCON->PIO0[n] = tmp;

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);
}

/*
 *
 */
void drv_select_method_GPIO(void)
{
	u_long n, tmp;

	/*
	 * de-assign SPI0_MOSI output by Switch-matrix
	 */
	n = 4;
	tmp = LPC_SWM->PINASSIGN[n];
	tmp &= ~(0xff <<0);
	tmp |=  (0xff <<0);
	LPC_SWM->PINASSIGN[n] = tmp;

	/*
	 * memory current choice
	 */
	drv_use_method_SPI = FALSE;
}

/*
 *
 */
void drv_select_method_SPI(void)
{
	u_long n, tmp;

	/*
	 * assign SPI0_MOSI to PIO0_3 by Switch matrix
	 */
	n = 4;
	tmp = LPC_SWM->PINASSIGN[n];
	tmp &= ~(0xff <<0);
	tmp |=  (PIO0_n_DIN <<0);
	LPC_SWM->PINASSIGN[n] = tmp;

	/*
	 * memory current choice
	 */
	drv_use_method_SPI = TRUE;
}

/*
 * a switcher
 */
void drv_send_code(u_long *m, int N_leds)
{
	if(drv_use_method_SPI) {
		drv_SPI_send_code(m, N_leds);
	} else {
		drv_GPIO_send_code(m, N_leds);
	}
}

/* =============================================================== */

/*
 * drive the port as GPIO by manual
 */
void drv_GPIO_send_code(u_long *m, int N_leds)
{
	u_long nn;
	int bit, tt1, tt2;

	while(N_leds-- > 0) {
		/*
		 * get colors for 1 LED
		 */
		nn = *m++;

		/*
		 * send 8+8+8 bit code for the LED
		 */
		for(bit=0; bit<24; bit++) {
			nn <<= 1;
			if(nn & (1<<24)) {
				/*
				 * send '1'
				 */
				tt1 = drv_GPIO_pulse_wide;
				tt2 = drv_GPIO_pulse_narrow;
			} else {
				/*
				 * send '0'
				 */
				tt1 = drv_GPIO_pulse_narrow;
				tt2 = drv_GPIO_pulse_wide;
			}

			/*
			 * make a pulse
			 */
			drv_GPIO_make_a_pulse(tt1, tt2, BITMAP_DIN, &LPC_GPIO_PORT->SET[0]);
		}
	}

	/*
	 * do not make Tpause
	 *   insert another job, don't call me continuously
	 */
}

/*
 * make pulse for 1bit
 */
#if 0
	/*
	 * C version
	 */
void drv_GPIO_make_a_pulse(int tt1, int tt2, int bitmap, VOLATILE u_long *reg)
{
	/*
	 * set High
	 */
	*reg = bitmap;

	do {
		asm("nop");			/* avoid elimination by optimizer */
	} while(--tt1 > 0);

	/*
	 * set Low
	 */
	reg += &LPC_GPIO_PORT->CLR[0] - &LPC_GPIO_PORT->SET[0];
	*reg = bitmap;		/* next Low */

	do {
		asm("nop");			/* avoid elimination by optimizer */
	} while(--tt2 > 0);
}
#else /* n */
	/*
	 * asm version
	 *   (int tt1, int tt2, int bitmap, VOLATILE u_long *reg);
	 *        r0       r1       r2                       r3
	 */
	asm(".thumb");
	asm(".section .text, \"ax\"");
	asm(".global drv_GPIO_make_a_pulse");
	asm(".func drv_GPIO_make_a_pulse");
	asm("drv_GPIO_make_a_pulse:");
	asm("	str		r2, [r3]");		/* SET[0] = bitmap */
	asm("lp01:");
	asm("	sub		r0, #1");		/* always update flags in THUMB2 */
//	asm("	cmp		r0, #0");
	asm("	bgt		lp01");
	asm("	add		r3, #128");		/* 0x80 */
	asm("	str		r2, [r3]");		/* CLR[0] = bitmap */
	asm("lp02:");
	asm("	sub		r1, #1");		/* always update flags in THUMB2 */
//	asm("	cmp		r1, #0");
	asm("	bgt		lp02");
	asm("	bx		lr");
	asm(".endfunc");
#endif /* n */

/* =============================================================== */

/*
 * make drive pulse like below by SPI peripheral with 3MHz clock
 *
 *     -->|   |<--      :  333nsec
 *     -->|       |<--  :  667nsec
 *        |<--------->| : 1000nsec
 *     -----------------------------> time
 *        +---+       +---+
 *        |   | 0     |   | 0        ... for "0 code"
 *      --+   +-------+   +-----
 *        +-------+   +-------+
 *        |     1 |   |     1 |      ... for "1 code"
 *      --+       +---+       +-
 *
 * use SPI as: SCK=3MHz
 *             15bits/frame                  ... 5usec/200kHz per frame
 *             send MSB-first
 *             shall use FIFO and interrupt  ... depth=8 on LPC1830
 *
 * divide 24bit-color-code to (5+5+5+5+4)bit with filling a '0' to final 1bit
 *
 *    |-----------|-----------|---------....------|-----------|-----------|
 *     G7 G6 G5 G4 G3 G2 G1 G0 R7 R6 R5 .... R1 R0 B7 B6 B5 B4 B3 B2 B1 B0 '0'
 *     ~~~~~~~~~~~~~~ ~~~~~~~~~~~~~~ ~~~    ~~~ ~~~~~~~~~~~~~~ ~~~~~~~~~~~~~~
 *      |    ...   |               expand to 15bit:
 *      |          +------------+         '1' to 1-1-0
 *      |                       |         '0' to 1-0-0
 *      V                       V               (ubfx instruction of thumb2)
 *   +-----+-----+-----+-----+-----+            (may be efficient)
 *    1 n 0 1 n 0 1 n 0 1 n 0 1 n 0
 *    -----|-------|-------|-------|  ===> set bits of 0x4924
 *                                                    (0x4920 when final)
 *
 * send these data continuously by SPI, final '0' makes:
 *   (1) force the MOSI output LOW after transfer
 *   (2) make a pause at every 24bit code (per device), like
 *
 *          ...|||| |||||||||||||||||||||||| |||||...
 *
 * finally, send NULL frame as 'reset-code'
 *   it takes 5usec per frame, 20usec is enough
 */

/*
 * about SPI frame conversion
 */
#define N_DIV_BITs         5			/* send 5bit at once */
#define N_DIV_COMBI  (1<<N_DIV_BITs)	/* number of combinations: 32 */
#define DIV_MASK     (N_DIV_COMBI-1)	/* bitmask for 5bit: 0x1f */
#define N_FRAMEs           5	/* it needs 5loop to send 24bit-color-value */

/*
 * about buffer for SPI-frame-data
 */
#define N_PAUSE     4	/* how many frames sent as reset-code [5usec/1] */

/*
 * constant table for 5 frames (by octal)
 */
static const u_short const_tab[N_FRAMEs] = {
	044444,		/* for G7,G6,G5,G4,G3 */
	044444,		/* for G2,G1,G0,R7,R6 */
	044444,		/* for R5,R4,R3,R2,R1 */
	044444,		/* for R0,B7,B6,B5,B4 */
	044440,		/* for B3,B2,B1,B0, 0 */
};

/*
 * convertion table for 5bit-input (by octal)
 */
static const u_short expand_tab[N_DIV_COMBI] = {
	000000, 000002, 000020, 000022,  000200, 000202, 000220, 000222,
	002000, 002002, 002020, 002022,  002200, 002202, 002220, 002222,
	020000, 020002, 020020, 020022,  020200, 020202, 020220, 020222,
	022000, 022002, 022020, 022022,  022200, 022202, 022220, 022222,
};
#if 00000000
	/*
	 * above table is made by AWK
	 *   (2^n) is identical to (1<<n)
	 */
	BEGIN {
		for(i=0; i<(2^5); i++) {
			nn = i;
			zz = 0x0000;
			if(nn >= (2^4)) { zz += (2^13); nn -= (2^4)}
			if(nn >= (2^3)) { zz += (2^10); nn -= (2^3)}
			if(nn >= (2^2)) { zz += (2^ 7); nn -= (2^2)}
			if(nn >= (2^1)) { zz += (2^ 4); nn -= (2^1)}
			if(nn >= (2^0)) { zz += (2^ 1); nn -= (2^0)}
			printf("0%05o,\n", zz);
		}
	}
#endif /* 00000000 */

/*
 * drive the port by SPI
 */
void drv_SPI_send_code(u_long *m, int N_leds)
{
	u_long nn, waveform;
	int i, bits;

	while(N_leds-- > 0) {
		/*
		 * get colors for 1 LED
		 */
		nn = *m++;

		/*
		 * send the code by 5bit x 5times
		 */
		for(i=0; i<N_FRAMEs; i++) {
			bits = (nn >> 19) & DIV_MASK;
			waveform = expand_tab[bits] | const_tab[i];

			/*
			 * confirm the FIFO is not full
			 */
			while(! (LPC_SPI0->STAT & SPI_STAT_TXRDY) ) {
				;
			}
			LPC_SPI0->TXDAT = waveform;

			/*
			 *
			 */
			nn <<= 5;
		}
	}

	/*
	 * make PAUSE interval
	 */
	for(i=0; i<N_PAUSE; i++) {
		/*
		 * confirm the FIFO is not full
		 */
		while(! (LPC_SPI0->STAT & SPI_STAT_TXRDY) ) {
			;
		}
		LPC_SPI0->TXDAT = 0x0000;
	}
}

/* =============================================================== */
