/*
 * main.c - main C entry
 *   by uratan! 2014.10.16
 */
//#define __MAIN_C__		/* <== inform common.h to alloc variables */
#include "common.h"
#include <chip.h>
	/*
	 * my LPC810 (DIP8)
	 *   +------------+
	 *   | LPC810     |
	 *   ) P  M021F   |
	 *   ) X  1M28603 |
	 *   | N  +D3414C | ... year 3, week 41, 4C' ver.13.4
	 *   +------------+
	 *      | Dev.ID:  0x00008100
	 *      | B.L.Ver: 13.4
	 */
/*
 * I/O pin usage
 *   pin#3: PIO0_3 (output) - drive WS2812B (GPIO or SPI)
 *   pin#4: PIO0_2 (input)  - U1_RxD from Ir receiver
 */

/*
 * use external clock thru CLKIN
 */
#define USE_CLKIN_12MHz  0

/*
 * SysTick base interval
 */
#define SYSTICK_Hz   10

/*
 * functions
 */
int main(void);
void configure_fixed_pins(void);
u_long initial_sysCLK(int use_PLL_or_set_divider);
#if 00000
u_long try_to_set_power(void);
#endif /* 00000 */
void init_SysTick(u_long sysCLK);
void SysTick_Handler(void)        ATTR_INTR;
void def_Exception_Handler(void)  ATTR_INTR;
void def_NMI_Handler(void)        ATTR_INTR;
void def_HardFault_Handler(void)  ATTR_INTR;
void def_SVCall_Handler(void)     ATTR_INTR;
void def_PendSV_Handler(void)     ATTR_INTR;
void get_specials(void);
void disp_specials(void);

/*
 * variables
 */
VOLATILE int systick_raw;
VOLATILE int systick_1Hz;

/*
 * externs
 */
/* in wconsole.c */
extern void wc_console(void);
/* in drive.c */
extern void drive_c_init(u_long sysCLK);
/* in uart1.c */
extern void uart1_c_init(u_long baudrate, u_long mainCLK);
/* in d_regs.c */
extern void dump_some_regs(void);


/* =============================================================== */

/*
 *
 */
int main(void)
{
	u_long tmp;
	u_long sysCLK;
	u_long mainCLK;

	/*
	 * mask all interrupts by PRIMASK flag in the core
	 */
	__disable_irq();

	/*
	 * initial my variables
	 */
	systick_raw = 0;
	systick_1Hz = 0;

	/*
	 * configure I/O pins
	 */
	configure_fixed_pins();

#if 1
	/*
	 * the flashROM may be faster enough to use it with no-wait even @ 30MHz...
	 *   (http://www.lpcware.com/content/forum/lpc800-speed-flash/)
	 */
	tmp = LPC_FMC->FLASHCFG;
	tmp = (tmp & ~0x3) | 0x0;	/* flash access time = 1 system clock */
	LPC_FMC->FLASHCFG = tmp;
#endif /* n */

#if USE_CLKIN_12MHz
	/*
	 * setup system clock for external 12MHz
	 */
	sysCLK = initial_sysCLK(/*use_PLL_or_set_divider=*/ 0);
#else /* USE_CLKIN_12MHz */
	/*
	 * setup PLL for max 30MHz processing speed
	 *   (this is for driving WS2812B thru GPIO by bit bang)
	 */
	sysCLK = initial_sysCLK(/*use_PLL_or_set_divider=*/ 1);
#endif /* USE_CLKIN_12MHz */
	mainCLK = sysCLK * LPC_SYSCTL->SYSAHBCLKDIV;

	/*
	 * setup uart for debug, etc.
	 */
	cm_init_uart(57600, mainCLK);

	/*
	 *
	 */
	rprintf(
		""							"\r\n"
		"/*"						"\r\n"
		" * hello...(%dMHz/%dMHz)"	"\r\n"
		" */"						"\r\n"
		""							"\r\n"
		, mainCLK / 1000 / 1000
		, sysCLK / 1000 / 1000
	);
	disp_specials();	/* sampled in Startup.S */
	dump_some_regs();
//rprintf("0x0c00_0140 is: %d\r\n",   *(u_long *)0x0c000140);
//rprintf("0x4004_8224 is: %08x\r\n", *(u_long *)0x40048224);
	rprintf("\r\n");

	/*
	 * initial peripherals
	 */
	drive_c_init(sysCLK);
	uart1_c_init(2400, mainCLK);
	init_SysTick(sysCLK);

	/*
	 * start my job
	 */
	__enable_irq();

	wc_console();

	/* never return */
}

/* =============================================================== */

/*
 * determine how handle fixed pins
 */
void configure_fixed_pins(void)
{
	u_long tmp;

	tmp = LPC_SWM->PINENABLE0;

	/*
	 * disable these fixed-pin-function, GPIO will have control
	 */
	tmp |= 0
			| (1<<2)		/* disable SWCLK, PIO0_3 enabled */
			| (1<<3)		/* disable SWDIO, PIO0_2 enabled */
		//  | (1<<6)		/* disable RESET, PIO0_5 enabled */
			| 0;

	/*
	 * enable these fixed-pin-function, GPIO will lose control
	 *   (do not enable both ACMP_I2 and CLKIN at same time)
	 */
	tmp &= ~(0
		//  | (1<<0)		/* enable ACMP_I1, PIO0_0 disabled */
		//  | (1<<1)		/* enable ACMP_I2, PIO0_1 disabled */
//NO-PIN//  | (1<<4)		/* enable XTALIN,  PIO0_8 disabled */
//NO-PIN//  | (1<<5)		/* enable XTALOUT, PIO0_9 disabled */
#if USE_CLKIN_12MHz
			| (1<<7)		/* enable CLKIN,   PIO0_1 disabled */
#else /* USE_CLKIN_12MHz */
		//  | (1<<7)		/* enable CLKIN,   PIO0_1 disabled */
#endif /* USE_CLKIN_12MHz */
//NO-PIN//  | (1<<8)		/* enable VDDCMP,  PIO0_6 disabled */
			| 0);

	/*
	 * write back
	 */
	LPC_SWM->PINENABLE0 = tmp;
}

/* =============================================================== */

/*
 * configure system clock
 *   (param > 0) start PLL to run at 30MHz
 *   (param ==0) run at IRC 12MHz
 *   (param < 0) run at IRC 12MHz with divider
 */
	/*
	 * set_pll() ROM-API can not configure 30MHz from 12MHz...
	 *   (he can configure only 1.5x, 2x, 3x of input frequency)
	 */
u_long initial_sysCLK(int use_PLL_or_set_divider)
{
	u_long divider;

	/*
	 * expect these conditions by cold start
	 */
//	LPC_SYSCTL->SYSAHBCLKDIV = 1;		/* 1/1 (reset value) */
//	LPC_SYSCTL->SYSPLLCLKSEL = 0;		/* sel IRC (reset value) */
//	LPC_SYSCTL->SYSPLLCLKUEN = 0;
//	LPC_SYSCTL->SYSPLLCLKUEN = 1;
//	LPC_SYSCTL->MAINCLKSEL = 0x0;		/* run with IRC (reset value) */
//	LPC_SYSCTL->MAINCLKUEN = 0;
//	LPC_SYSCTL->MAINCLKUEN = 1;

#if USE_CLKIN_12MHz
	/*
	 * use external 12MHz thru CLKIN
	 */
	LPC_SYSCTL->SYSPLLCLKSEL = 0x3;		/* CLKIN */
	LPC_SYSCTL->SYSPLLCLKUEN = 0;
	LPC_SYSCTL->SYSPLLCLKUEN = 1;
	LPC_SYSCTL->MAINCLKSEL = 0x1;		/* PLL clock-in */
	LPC_SYSCTL->MAINCLKUEN = 0;
	LPC_SYSCTL->MAINCLKUEN = 1;
#endif /* USE_CLKIN_12MHz */

	if(use_PLL_or_set_divider > 0) {
		/*
		 * start PLL for mainCLK=60MHz (12MHz * 5 = 240MHz / (2 * 2))
		 */
		LPC_SYSCTL->SYSPLLCTRL = (/*PSEL=*/ 0x01 <<5) | (/*MSEL=*/ 5 -1);
		LPC_SYSCTL->PDRUNCFG &= ~(1<<7);		/* enable PLL */
		while(! (LPC_SYSCTL->SYSPLLSTAT & 1) ) {
			;	/* wait for lock */
		}

		/*
		 * connect PLL output
		 */
		LPC_SYSCTL->SYSAHBCLKDIV = 2;		/* 1/2 */
		LPC_SYSCTL->MAINCLKSEL = 0x3;		/* select PLL-output */
		LPC_SYSCTL->MAINCLKUEN = 0;
		LPC_SYSCTL->MAINCLKUEN = 1;

		return(30*1000*1000);	/* sysCLK is: 30MHz */
	}
	else {
		divider = - use_PLL_or_set_divider;
		if(divider == 0) {
			divider = 1;
		} else if(divider > 255) {
			divider = 255;
		}
		LPC_SYSCTL->SYSAHBCLKDIV = divider;
		return(12*1000*1000 / divider);		/* return sysCLK */
	}
}

#if 00000
/*
 * examine set_power() ROM-API
 */
u_long try_to_set_power(void)
{
	uint32_t cmd[4];
	uint32_t resp[2];
	u_long sysCLK_MHz;

//	/*
//	 * try set_power tuning
//	 */
//	cmd[0] = sysCLK_MHz;		/* [MHz] */
//	cmd[1] = PWR_DEFAULT;
//	cmd[2] = sysCLK_MHz;		/* [MHz] */
//	LPC_PWRD_API->set_power(cmd, resp);

	return(resp[0]);
}
#endif /* 00000 */

/* =============================================================== */

/*
 * run SysTick timer
 */
void init_SysTick(u_long sysCLK)
{
	/*
	 * setup SysTick timer
	 */
	SysTick->LOAD = sysCLK / SYSTICK_Hz / 2 - 1;	/* use half-clock */
	SysTick->VAL = 0;			/* clear the counter (value is ignored) */
		/*
		 * 30,000,000 = 0x01C9_C380  >  (1<<24)
		 * 16,777,215 = 0x00FF_FFFF  =  (1<<24)
		 * 15,000,000 = 0x00E4_E1C0  <  (1<<24) ... half-clock covers 1Hz
		 */
	/*
	 * setup NVIC
	 */
	NVIC_SetPriority(SysTick_IRQn, PRI_SYSTICK);

	/*
	 * enable intr and run it
	 */
	SysTick->CTRL = 0x03;		/* run it with half-clock, intr enabled */
}

/*
 * interrupt service routine
 */
void SysTick_Handler(void)
{
	static int divider = 0;

	systick_raw = 1;

	if(++divider >= SYSTICK_Hz) {
		divider = 0;
		systick_1Hz = 1;
	}
		/*
		 * interrupt request flag may be cleared automatically
		 */
}

/* =============================================================== */

/*
 * try to catch exceptions
 */

void def_Exception_Handler(void)
{
	rprintf("! some Exception\r\n");
	while(1)
		;
	/* never return */
}

void def_NMI_Handler(void)
{
	rprintf("! NMI\r\n");
	while(1)
		;
	/* never return */
}

void def_HardFault_Handler(void)
{
	rprintf("! HardFault\r\n");
	while(1)
		;
	/* never return */
}

void def_SVCall_Handler(void)
{
	rprintf("! SVCall\r\n");
	while(1)
		;
	/* never return */
}

void def_PendSV_Handler(void)
{
	rprintf("! PendSV\r\n");
	while(1)
		;
	/* never return */
}

/* =============================================================== */

/*
 * try to watch Cortex-M0+ special registers
 */
	/*
	 *  (*) read EPSR returns always 0x00
	 */
u_long m0p_PSP;			/* Process Stack Pointer */
u_long m0p_MSP;			/* Main Stack Pointer */
u_long m0p_PSR;			/* Program status register */
//u_long m0p_APSR;			/*  Application PSR */
//u_long m0p_IPSR;			/*  Interrupt PSR */
//u_long m0p_EPSR;			/*  Excution PSR */
u_long m0p_PRIMASK;		/* Priority Mask */
u_long m0p_CONTROL;		/* Control */

/*
 * get contents of special registers
 */
void get_specials(void)
{
	/*
	 *   MRS --- Moveto-reg-from-SpecialReg
	 *         { APSR, IPSR, EPSR,           ... single
	 *           IEPSR, IAPSR, EAPSR,        ... dual OR'ed
	 *           PSR,                        ... all OR'ed ?
	 *           PRIMASK,                    ... what's this ?
	 *           CONTROL, MSP, PSP
	 *         }
	 */
	asm("	.extern m0p_PSP");
	asm("	LDR		R1, =m0p_PSP");
	asm("	MRS		R2, PSP");
	asm("	STR		R2, [R1]");

	asm("	.extern m0p_MSP");
	asm("	LDR		R1, =m0p_MSP");
	asm("	MRS		R2, MSP");
	asm("	STR		R2, [R1]");

	asm("	.extern m0p_PSR");
	asm("	LDR		R1, =m0p_PSR");
	asm("	MRS		R2, PSR");
	asm("	STR		R2, [R1]");

//	asm("	.extern m0p_APSR");
//	asm("	LDR		R1, =m0p_APSR");
//	asm("	MRS		R2, APSR");
//	asm("	STR		R2, [R1]");
//
//	asm("	.extern m0p_IPSR");
//	asm("	LDR		R1, =m0p_IPSR");
//	asm("	MRS		R2, IPSR");
//	asm("	STR		R2, [R1]");
//
//	asm("	.extern m0p_EPSR");
//	asm("	LDR		R1, =m0p_EPSR");
//	asm("	MRS		R2, EPSR");
//	asm("	STR		R2, [R1]");

	asm("	.extern m0p_PRIMASK");
	asm("	LDR		R1, =m0p_PRIMASK");
	asm("	MRS		R2, PRIMASK");
	asm("	STR		R2, [R1]");

	asm("	.extern m0p_CONTROL");
	asm("	LDR		R1, =m0p_CONTROL");
	asm("	MRS		R2, CONTROL");
//asm("	LDR		R2, =0x123");		/* to confirm the code */
	asm("	STR		R2, [R1]");
}

/*
 * dump special registers
 */
void disp_specials(void)
{
	/*
	 * special registers
	 *           +----+----+----+----++----+----+----+----+
	 *           |3322|2222|2222|1111||1111|11  |    |    |
	 *           |1098|7654|3210|9876||5432|1098|7654|3210|
	 *           +----+----+----+----++----+----+----+----+
	 * APSR      |NZCV|    |    |    ||    |    |    |    |
	 *           +----+----+----+----++----+----+----+----+
	 * IPSR      |    |    |    |    ||    |    |  ISR_NUM|
	 *           +----+----+----+----++----+----+----+----+
	 * EPSR      |    |   T|    |    ||    |  a |    |    | T: thumb
	 *           +----+----+----+----++----+----+----+----+ a: (stack alignment)
	 * PRIMASK   |    |    |    |    ||    |    |    |   M| M: mask
	 *           +----+----+----+----++----+----+----+----+
	 * CONTROL   |    |    |    |    ||    |    |    |  PU| P: use PSP
	 *           +----+----+----+----++----+----+----+----+ U: Unprivileged
	 */
	rprintf(" m0p_PSP:      0x%08x\r\n", m0p_PSP);
	rprintf(" m0p_MSP:      0x%08x\r\n", m0p_MSP);
	rprintf(" m0p_PSR:      0x%08x\r\n", m0p_PSR);
//	rprintf("  m0p_APSR:      %08x\r\n", m0p_APSR);
//	rprintf("  m0p_IPSR:      %08x\r\n", m0p_IPSR);
//	rprintf("  m0p_EPSR:      %08x\r\n", m0p_EPSR);
	rprintf(" m0p_PRIMASK:  0x%08x\r\n", m0p_PRIMASK);
	rprintf(" m0p_CONTROL:  0x%08x\r\n", m0p_CONTROL);
}
	/*
	 * --------------------------------------
	 * 'processor mode' and 'privilege level'
	 * --------------------------------------
	 * RESET:
	 *     IPSR      = 0x0000_0000;
	 *     PRIMASK   = 0x0000_0000;     // not masked
	 *     CONTROL   = 0x0000_0000;     // MSP, Privileged
	 *     goto Entry;
	 *
	 * if(Handler_mode) {
	 *     //// exceptions, interrupts
	 *     SP         = MSP;		// MainSP  ( <=> ProcessSP)
	 *     Privileged = yes;
	 * } else {
	 *   Entry:
	 *     //// Thread_mode (normal)
	 *     SP         = (CONTROL & 2) ? PSP : MSP;
	 *     Privileged = (CONTROL & 1) ? no  : yes;
	 * }
	 * --------------------------------------
	 *
	 *   interrupt mask state at start up time
	 *   may be guaranteed by NVIC unit ?
	 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

		/* there are a lot of params/states/APIs, grab them */
	/* ============================================ in core_cmFunc.h
	 *   void     __enable_irq(void);                  --- CPSIE I
	 *   void     __disable_irq(void);                 --- CPSID I
	 *   uint32_t __get_CONTROL(void);                 --- MRs ...
	 *   void     __set_CONTROL(uint32_t control);     --- mSR ...
	 *   uint32_t __get_IPSR(void);                    --- MRs ...
	 *   uint32_t __get_APSR(void);                    --- MRs ...
	 *   uint32_t __get_xPSR(void);                    --- MRs ...
	 *   uint32_t __get_PSP(void);                     --- MRs ...
	 *   void     __set_PSP(uint32_t topOfProcStack);  --- mSR ...
	 *   uint32_t __get_MSP(void);                     --- MRs ...
	 *   void     __set_MSP(uint32_t topOfMainStack);  --- mSR ...
	 *   uint32_t __get_PRIMASK(void);                 --- MRs ...
	 *   void     __set_PRIMASK(uint32_t priMask);     --- mSR ...
	 */
	/* ============================================ in core_cm0plus.h
	 *   void     NVIC_EnableIRQ(IRQn_Type IRQn);            --- NVIC->ISER[]
	 *   void     NVIC_DisableIRQ(IRQn_Type IRQn);           --- NVIC->ICER[]
	 *   uint32_t NVIC_GetPendingIRQ(IRQn_Type IRQn);        --- NVIC->ISPR[]
	 *   void     NVIC_SetPendingIRQ(IRQn_Type IRQn);        --- NVIC->ISPR[]
	 *   void     NVIC_ClearPendingIRQ(IRQn_Type IRQn);      --- NVIC->ICPR[]
	 *   void     NVIC_SetPriority(IRQn_Type n, uint32_t p); --- SCB->SHP[] or NVIC->IP[]
	 *   uint32_t NVIC_GetPriority(IRQn_Type IRQn);          --- SCB->SHP[] or NVIC->IP[]
	 * o void     NVIC_SystemReset(void);                    --- SCB->AIRCR
	 */

/* =============================================================== */
