/*
 * wconsole.c - a console for WS2812B
 *   by uratan! 2014.10.18
 */
#include "common.h"
#include <chip.h>

/*
 * version of this firmware
 */
#define VERSION  "0.21"

/*
 * how many LEDs connected
 */
#define N_LEDs   3
//#define N_LEDs   60

/*
 * typedef for brightness values
 *   |<--------- long ---------->|
 *   +------+------+------+------+
 *   | dumm | green| red  | blue |
 *   +------+------+------+------+
 */
typedef union {
	u_long grb;
	struct {
		u_char b;		/* because his endian is little */
		u_char r;
		u_char g;
		u_char dumm;
	} c;
} T_CODE_GRB;

/*
 * functions
 */
void wc_console(void);
void wc_help(void);
void wc_tune_color_value(u_char *var, int pol, int key);
void wc_tune_pulse_width(int *var, int delta);

/*
 * variables
 */
int mode_auto;
T_CODE_GRB code_grb[N_LEDs];

/*
 * externs
 */
/* in main.c */
extern VOLATILE int systick_1Hz;
/* in drvie.c */
extern void drv_select_method_GPIO(void);
extern void drv_select_method_SPI(void);
extern void drv_send_code(u_long *m, int N_leds);
extern int drv_use_method_SPI;
extern int drv_GPIO_pulse_wide;
extern int drv_GPIO_pulse_narrow;
/* in d_regs.c */
extern void dump_some_regs(void);

/*
 *
 */
void wc_console(void)
{
	int i, c;
	u_long tmp;
	int do_send;

	/*
	 * initial mine
	 */
	mode_auto = 0;
	for(i=0; i<N_LEDs; i++) {
		code_grb[i].grb = i & 0x03;
	}
	code_grb[0].grb = 0x0f0000;
	code_grb[1].grb = 0x000f00;
	code_grb[2].grb = 0x00000f;

	/*
	 * start console
	 */
	wc_help();

	/*
	 * loop forever
	 */
	while(1) {
		/*
		 * display prompt
		 */
		cm_putchar(']');

		/*
		 * wait for events
		 */
		while(1) {
			i = cm_keyin();
			if(i != -1) {
				c = i;
				break;		/* something arrived */
			}

			if(systick_1Hz) {
				systick_1Hz = 0;

				if(mode_auto) {
					c = KEY_CR;
					break;
				}
			}
		}

		/*
		 * parse the input
		 */
		do_send = 0;

		switch(c) {
		case 's':
				rprintf("SPI\r\n");
				drv_select_method_SPI();
				break;

		case 'g':
				rprintf("GPIO\r\n");
				drv_select_method_GPIO();
				break;

		case 't':
				mode_auto = 1 - mode_auto;
				rprintf("auto-mode: %s\r\n", (mode_auto) ? "ON" : "off");
				break;

		case KEY_CR:
				rprintf("rotate FWD\r\n");
				tmp = code_grb[0].grb;
				for(i=0; i<(N_LEDs-1); i++) {
					code_grb[i].grb = code_grb[i+1].grb;
				}
				code_grb[i].grb = tmp;
				do_send = 1;
				break;

		case KEY_BS:
				rprintf("rotate RVS\r\n");
				tmp = code_grb[N_LEDs-1].grb;
				for(i=(N_LEDs-1); i>0; i--) {
					code_grb[i].grb = code_grb[i-1].grb;
				}
				code_grb[0].grb = tmp;
				do_send = 1;
				break;

		case KEY_SPC:
				do_send = 1;
				break;
#if 0000
		case KEY_ESC:
				rprintf("continuous send\r\n");
				while(1) {
					drv_send_code(&code_grb[0].grb, N_LEDs);
				}
				break;
#endif /* 0000 */

		case 'u':
		case 'U':
				wc_tune_color_value(&code_grb[0].c.g,  1, c);
				do_send = 1;
				break;
		case 'j':
		case 'J':
				wc_tune_color_value(&code_grb[0].c.g, -1, c);
				do_send = 1;
				break;
		case 'i':
		case 'I':
				wc_tune_color_value(&code_grb[0].c.r,  1, c);
				do_send = 1;
				break;
		case 'k':
		case 'K':
				wc_tune_color_value(&code_grb[0].c.r, -1, c);
				do_send = 1;
				break;
		case 'o':
		case 'O':
				wc_tune_color_value(&code_grb[0].c.b,  1, c);
				do_send = 1;
				break;
		case 'l':
		case 'L':
				wc_tune_color_value(&code_grb[0].c.b, -1, c);
				do_send = 1;
				break;

		case 'n':
				wc_tune_pulse_width(&drv_GPIO_pulse_narrow, -1);
				break;
		case 'm':
				wc_tune_pulse_width(&drv_GPIO_pulse_narrow,  1);
				break;
		case ',':
				wc_tune_pulse_width(&drv_GPIO_pulse_wide, -1);
				break;
		case '.':
				wc_tune_pulse_width(&drv_GPIO_pulse_wide,  1);
				break;

		case 'd':
				dump_some_regs();
				break;
#if 000
		case 'b':
#if 0
# define DUMP_START  0x1fff0000	/* dump bootROM (8kB) */
# define DUMP_END    0x1fff1fff
#else /* n */
# define DUMP_START  0x0c000000	/* dump ??? (512B) */
# define DUMP_END    0x0c0001ff
#endif /* n */
		{			/* dump something */
				int iii;
				u_char *bb;

				bb = (u_char *)DUMP_START;
				while(1) {
					if(bb > (u_char *)DUMP_END) {
						break;
					}
					rprintf("%08x:", (u_long)bb);
					for(iii=0; iii<16; iii++) {
						rprintf(" %02x", *bb++);
					}
					rprintf("\r\n");
				}
		}
				break;
#endif /* 000 */

		case '?':
		default:
				wc_help();
				break;
		}

		if(do_send) {
			rprintf("! %06x-%06x-...\r\n", code_grb[0].grb, code_grb[1].grb);
			drv_send_code(&code_grb[0].grb, N_LEDs);
		}
	}

	/* never return */
}

/*
 *
 */
void wc_help(void)
{
	rprintf(
		""												"\r\n"
		"/*"											"\r\n"
		" * WS2812B bench for LPC810 ver." VERSION		"\r\n"
		" *  [command list]  (N_LEDs is: %d)"			"\r\n"
		" *   s,g -- select drive method: SPI/GPIO"		"\r\n"
		" *   t -- toggle auto-mode: ON/off"			"\r\n"
		" *   RET,BS -- rotate values"					"\r\n"
		" *   u,i,o -- inc G,R,B"						"\r\n"
		" *   j,k,l -- dec G,R,B"						"\r\n"
		" *   SPC -- send"								"\r\n"
		" *   n/m -- tune GPIO narrow pulse width"		"\r\n"
		" *   ,/. -- tune GPIO  wide  pulse width"		"\r\n"
		" *   d --- dump LPC810 regs"					"\r\n"
		" *   ? --- display this menu"					"\r\n"
		" */"											"\r\n"
		, N_LEDs
	);
	if(drv_use_method_SPI) {
		rprintf(" SPI mode\r\n");
	} else {
		rprintf(" GPIO mode, (n:%d,w:%d)\r\n", drv_GPIO_pulse_narrow, drv_GPIO_pulse_wide);
	}
}

/*
 * inc/dec G,R,B element
 */
void wc_tune_color_value(u_char *var, int pol, int key)
{
	int val;

	val = *var;

	if(key & 0x20) {
		/* lower case */
		val += pol * 1;
	} else {
		/* upper case */
		val += pol * 16;
	}

	if(val < 0) {
		val = 0;
	} else if(val > 255) {
		val = 255;
	}

	*var = val & 0xff;

	rprintf("  %06x\r\n", code_grb[0]);
}

/*
 *
 */
void wc_tune_pulse_width(int *var, int delta)
{
	int val;

	val = *var;
	val += delta;
	if(val < 0)
		val = 0;
	else if(val > 1000)
		val = 1000;
	*var = val;

	rprintf(" (n:%d,w:%d)\r\n", drv_GPIO_pulse_narrow, drv_GPIO_pulse_wide);
}
